#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.interfaces.task.handler module

This module defines interfaces used by file handlers.
"""

__docformat__ = 'restructuredtext'

from zope.interface import Attribute, Interface, Invalid, implementer, invariant
from zope.schema import Bool, Int, Object, Password, TextLine, Tuple
from zope.schema.interfaces import ITuple
from pyams_scheduler.interfaces import ITask

from pyams_scheduler import _


class IDirectoryHandlerHostField(ITuple):
    """Handler host field marker interface"""


@implementer(IDirectoryHandlerHostField)
class DirectoryHandlerHostField(Tuple):
    """Directory handler host field

    This schema field is made of two elements which are the protocol
    (file, ftp, sftp...) and the host (which can be None for local files
    handlers).
    """

    def __init__(self, **kwargs):
        super().__init__(value_type=TextLine(),
                         min_length=2,
                         max_length=2,
                         **kwargs)


class IDirectoryHandler(Interface):
    """Directory handler interface

    This interface tries to mimic the *os* package files interfaces
    required to handle directory synchronization.
    """

    directory = Attribute("Current directory")

    def connect(self):
        """Open connection if required"""

    def close(self):
        """Close handler connection"""

    def open(self, name, mode='r'):
        """Open file

        This method may be a context manager so that connection will
        be closed automatically on exit.
        """

    def chdir(self, dirname):
        """Change current directory

        This method may be a context manager, so that previous directory
        is automatically restored on exit.
        """

    def mkdir(self, dirname):
        """Create new directory"""

    def listdir(self):
        """List current directory content"""

    def stat(self, name):
        """Get stats about given file name"""

    def set_time(self, name, atime, mtime):
        """Set atime and mtime of given file name"""

    def set_mode(self, name, mode=0x777):
        """Set mode of given file name"""

    def unlink(self, name):
        """Delete given file name"""


class IDirectoryInfo(Interface):
    """Synchronization task connection information"""

    host = DirectoryHandlerHostField(title=_("Host"),
                                     description=_("Protocol and host name"),
                                     required=True)

    port = Int(title=_("Port number"),
               description=_("Port number is required for remote protocols like SFTP or FTP"),
               default=22,
               required=False)

    auto_add_host_key = Bool(title=_("Automatically add host key?"),
                             description=_("If 'yes', an SSH connection will be accepted "
                                           "even if server key is unknown"),
                             required=False,
                             default=False)

    username = TextLine(title=_("User name"),
                        required=False)

    private_key = TextLine(title=_("Private key filename"),
                           description=_("Enter name of private key file; use '~' to identify "
                                         "running server user home directory, as in "
                                         "~/.ssh/id_rsa"),
                           required=False)

    password = Password(title=_("Password"),
                        description=_("If not using private key, you must provide user's "
                                      "password"),
                        required=False)

    directory = TextLine(title=_("Directory name"),
                         description=_("Enter absolute directory name"),
                         required=True)

    @invariant
    def check_remote_host(self):
        """Check for password or private key if remote host is defined"""
        if self.host and self.host[1] and (bool(self.private_key) == bool(self.password)):  # pylint: disable=unsubscriptable-object
            raise Invalid(_("You must provide a private key filename OR a password when "
                            "defining remote tasks"))

    def get_handler(self):
        """Get handler matching this directory"""


class IDirectorySyncTaskInfo(Interface):
    """Directory synchronization task info"""

    source = Object(title=_("Source directory"),
                    schema=IDirectoryInfo,
                    required=True)

    filter = TextLine(title=_("Files names filter"),
                      description=_("Filter used to select only a subset of directory "
                                    "contents"),
                      required=False,
                      default='*')

    recursive = Bool(title=_("Recursive copy?"),
                     description=_("If 'yes', sub-directories will also be synchronized"),
                     required=False,
                     default=True)

    follow_symlinks = Bool(title=_("Follow symbolic links?"),
                           description=_("If 'yes', symbolic links targets will be copied as "
                                         "normal files; be warmed that symbolic links can be "
                                         "seen as regular files while using an SFTP source!"),
                           required=False,
                           default=False)

    ignore_read_errors = Bool(title=_("Ignore read errors?"),
                              description=_("If 'yes', source files which can't be opened "
                                            "will just be ignored; otherwise, read errors "
                                            "will stop the synchronization"),
                              required=False,
                              default=False)

    use_datetime = Bool(title=_("Use datetime?"),
                        description=_("If 'yes', last modification time will be used to "
                                      "compare files between source and target, and last "
                                      "modification date will be restored on target after copy"),
                        required=False,
                        default=True)

    delete_source = Bool(title=_("Delete source after copy"),
                         description=_("If 'yes', source directory will be cleared after a "
                                       "successful synchronization"),
                         required=False,
                         default=False)

    ignore_delete_errors = Bool(title=_("Ignore delete errors?"),
                                description=_("If 'yes', source files which can't be deleted "
                                              "after copy will just be ignored; otherwise, "
                                              "delete errors will stop the synchronization"),
                                required=False,
                                default=False)

    target = Object(title=_("Target directory"),
                    schema=IDirectoryInfo,
                    required=True)

    ignore_write_errors = Bool(title=_("Ignore write errors?"),
                               description=_("If 'yes', target files which can't be opened "
                                             "will just be ignored; otherwise, write errors "
                                             "will stop the synchronization"),
                               required=False,
                               default=False)


class IDirectorySyncTask(ITask, IDirectorySyncTaskInfo):
    """Directory synchronization task interface"""


class DirectorySyncError(Exception):
    """Base directory synchronization error"""


class DirectoryReadError(DirectorySyncError):
    """Directory read error"""


class DirectoryWriteError(DirectorySyncError):
    """Directory write error"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.history module

This module defines views which are used to display content history.
"""

from datetime import timedelta

from zope.interface import Interface
from zope.traversing.interfaces import ITraversable

from pyams_form.field import Fields
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_scheduler.interfaces import IScheduler, ITask, ITaskHistory, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task import TASK_STATUS_STYLES
from pyams_scheduler.task.zmi import TaskBaseFormMixin
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_skin.interfaces.viewlet import IContentPrefixViewletManager
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextAdapter, ContextRequestViewAdapter, adapter_config
from pyams_utils.date import SH_DATETIME_FORMAT, get_duration
from pyams_utils.traversing import get_parent
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalDisplayForm
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IPropertiesMenu
from pyams_zmi.table import ActionColumn, DateColumn, I18nColumnMixin, InnerTableAdminView, \
    NameColumn, Table, TableAdminView, TableElementEditor
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'


from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@viewlet_config(name='jobs-history.menu',
                context=IScheduler, layer=IAdminLayer,
                manager=IPropertiesMenu, weight=30,
                permission=MANAGE_TASKS_PERMISSION)
class SchedulerHistoryMenu(NavigationMenuItem):
    """Scheduler history menu"""

    label = _("Jobs history")
    href = '#jobs-history.html'


class SchedulerHistoryTable(Table):
    """Scheduler history table"""

    css_classes = {
        'table': 'table table-striped table-hover table-xs datatable'
    }

    @property
    def data_attributes(self):
        attributes = super().data_attributes
        attributes['table'].update({
            'data-ams-order': '1,desc'
        })
        return attributes

    def get_css_highlight_class(self, column, item, css_class):
        return (css_class or '') + ' ' + TASK_STATUS_STYLES.get(item.status, 'table-warning')


@adapter_config(required=(IScheduler, IAdminLayer, SchedulerHistoryTable),
                provides=IValues)
class SchedulerHistoryTableValues(ContextRequestViewAdapter):
    """Scheduler history table values adapter"""

    @property
    def values(self):
        """Scheduler history table values getter"""
        for task in self.context.values():
            yield from task.history.values()


@adapter_config(name='name',
                required=(Interface, IAdminLayer, SchedulerHistoryTable),
                provides=IColumn)
class SchedulerHistoryNameColumn(NameColumn):
    """Scheduler history name column"""

    i18n_header = _("Task name")

    weight = 10

    def get_value(self, obj):
        task = get_parent(obj, ITask)
        return task.name


@adapter_config(name='date',
                required=(Interface, IAdminLayer, SchedulerHistoryTable),
                provides=IColumn)
class SchedulerHistoryDateColumn(I18nColumnMixin, DateColumn):
    """Scheduler history date column"""

    i18n_header = _("Run date")
    attr_name = 'date'
    formatter = SH_DATETIME_FORMAT

    weight = 20


@adapter_config(name='duration',
                required=(Interface, IAdminLayer, SchedulerHistoryTable),
                provides=IColumn)
class SchedulerHistoryDurationColumn(I18nColumnMixin, GetAttrColumn):
    """Scheduler history duration column"""

    i18n_header = _("Duration")
    attr_name = 'duration'

    weight = 30

    def get_value(self, obj):
        duration = super().get_value(obj)
        if duration == 0.:
            return '--'
        return get_duration(timedelta(seconds=duration), request=self.request)


@adapter_config(name='status',
                required=(Interface, IAdminLayer, SchedulerHistoryTable),
                provides=IColumn)
class SchedulerHistoryStatusColumn(I18nColumnMixin, GetAttrColumn):
    """Scheduler history status column"""

    i18n_header = _("Status")
    attr_name = 'status'

    weight = 40


@pagelet_config(name='jobs-history.html',
                context=IScheduler, layer=IPyAMSLayer,
                permission=MANAGE_TASKS_PERMISSION)
class SchedulerHistoryView(TableAdminView):
    """Scheduler history view"""

    title = _("Tasks execution history")
    table_class = SchedulerHistoryTable
    table_label = _("List of executed tasks")


#
# Task history view
#

@adapter_config(name='history',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskHistoryColumn(ActionColumn):
    """Scheduler task history column"""

    href = 'jobs-history.html'
    icon_class = 'fas fa-history'
    hint = _("Task run history")

    permission = MANAGE_TASKS_PERMISSION

    weight = 70


@pagelet_config(name='jobs-history.html', context=ITask, layer=IPyAMSLayer,
                permission=MANAGE_TASKS_PERMISSION)
class TaskHistoryView(TaskBaseFormMixin, AdminModalDisplayForm):
    """Task history view"""

    modal_class = 'modal-xl'


@viewlet_config(name='jobs-history-table',
                context=ITask, layer=IAdminLayer, view=TaskHistoryView,
                manager=IContentPrefixViewletManager, weight=10)
class TaskHistoryTableView(InnerTableAdminView):
    """Task history table view"""

    table_class = SchedulerHistoryTable
    table_label = _("List of executed jobs")


@adapter_config(required=(ITask, IAdminLayer, SchedulerHistoryTable),
                provides=IValues)
class TaskHistoryTableValues(ContextRequestViewAdapter):
    """Task history table values adapter"""

    @property
    def values(self):
        """Task history table values getter"""
        yield from self.context.history.values()


@adapter_config(name='history', required=ITask, provides=ITraversable)
class TaskHistoryTraverser(ContextAdapter):
    """Task history traverser"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Task history traverser"""
        return self.context.history


@adapter_config(required=(ITaskHistory, IAdminLayer, Interface),
                provides=ITableElementEditor)
class TaskHistoryElementEditor(TableElementEditor):
    """Task history element editor"""

    view_name = 'history.html'


@pagelet_config(name='history.html', context=ITaskHistory, layer=IPyAMSLayer)
class JobHistoryView(AdminModalDisplayForm):
    """Job run history display form"""

    @property
    def title(self):
        """Title getter"""
        return get_parent(self.context, ITask).name

    legend = _("Task run history")
    modal_class = 'modal-max'

    label_css_class = 'col-sm-3 col-md-2'
    input_css_class = 'col-sm-9 col-md-10'

    fields = Fields(ITaskHistory).omit('__parent__', '__name__')

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        duration = self.widgets.get('duration')
        if duration is not None:
            duration.value = get_duration(timedelta(seconds=self.context.duration))
        report = self.widgets.get('report')
        if report is not None:
            report.rows = 15
            report.add_class('monospace')



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.ssh module

This module defines a scheduler task which can be used to run local or
remote commands.
"""

import os
import subprocess
import sys
import traceback
from contextlib import contextmanager
from socket import gethostname

from paramiko import AutoAddPolicy, SSHClient, SSHException
from persistent import Persistent
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces.task import TASK_STATUS_ERROR, TASK_STATUS_OK
from pyams_scheduler.interfaces.task.ssh import ISSHCallerTask, ISSHConnectionInfo
from pyams_scheduler.task import Task
from pyams_utils.factory import factory_config


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@factory_config(ISSHConnectionInfo)
class SSHConnectionInfo(Persistent):
    """SFTP base info"""

    hostname = FieldProperty(ISSHConnectionInfo['hostname'])
    auto_add_host_key = FieldProperty(ISSHConnectionInfo['auto_add_host_key'])
    port = FieldProperty(ISSHConnectionInfo['port'])
    username = FieldProperty(ISSHConnectionInfo['username'])
    private_key = FieldProperty(ISSHConnectionInfo['private_key'])
    _password = FieldProperty(ISSHConnectionInfo['password'])

    def __init__(self, data=None):  # pylint: disable=unused-argument
        super().__init__()

    def __bool__(self):
        return bool(self.hostname)

    def __repr__(self):
        if self.hostname:
            return '{}@{}:{}'.format(self.username, self.hostname, self.port)
        return gethostname()

    @property
    def password(self):
        """Password getter"""
        return self._password

    @password.setter
    def password(self, value):
        """Password setter"""
        if value == '*****':
            return
        self._password = value

    @contextmanager
    def get_connection(self):
        """Open SSH connection"""
        ssh = SSHClient()
        try:
            if self.auto_add_host_key:
                ssh.set_missing_host_key_policy(AutoAddPolicy())
            ssh.connect(self.hostname, self.port, self.username, self.password,
                        key_filename=os.path.expanduser(self.private_key)
                        if self.private_key else None)
            yield ssh
        finally:
            ssh.close()

    @contextmanager
    def get_sftp_client(self):
        """Get SFTP client"""
        with self.get_connection() as ssh_client:
            try:
                yield ssh_client.open_sftp()
            finally:
                ssh_client.close()


@factory_config(ISSHCallerTask)
class SSHCallerTask(Task):
    """SSH command caller task"""

    label = _("Command line")
    icon_class = 'fas fa-terminal'

    connection = FieldProperty(ISSHCallerTask['connection'])
    cmdline = FieldProperty(ISSHCallerTask['cmdline'])
    ok_status = FieldProperty(ISSHCallerTask['ok_status'])

    @property
    def ok_status_list(self):
        """OK exit codes list getter"""
        return map(int, self.ok_status.split(','))

    def run(self, report, **kwargs):  # pylint: disable=unused-argument
        report.write('Shell command output\n'
                     '====================\n'
                     'Shell command: \n    {!r}: {}\n\n'.format(self.connection, self.cmdline))
        if self.connection:
            return self._run_remote(report, **kwargs)
        return self._run_local(report, **kwargs)

    def _run_remote(self, report, **kwargs):  # pylint: disable=unused-argument
        """Run remote SSH command"""
        try:
            with self.connection.get_connection() as ssh_client:
                stdin, stdout, stderr = ssh_client.exec_command(self.cmdline)
                stdin.close()
                report.write(stdout.read().decode())
                errors = stderr.read()
                if errors:
                    report.write('\n\n'
                                 'Some errors occurred\n'
                                 '====================\n')
                    report.write(errors.decode())
                return (
                    TASK_STATUS_OK if stdout.channel.exit_status in self.ok_status_list
                        else stdout.channel.exit_status,
                    stdout
                )
        except SSHException:
            etype, value, tb = sys.exc_info()  # pylint: disable=invalid-name
            report.write('\n\n'
                         'An SSH error occurred\n'
                         '=====================\n')
            report.write(''.join(traceback.format_exception(etype, value, tb)))
            return TASK_STATUS_ERROR, None

    def _run_local(self, report, **kwargs):  # pylint: disable=unused-argument
        """Run local system command"""
        shell = subprocess.Popen(self.cmdline, shell=True, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        stdout, stderr = shell.communicate()
        report.write(stdout.decode())
        if stderr:
            report.write('\n\n'
                         'Some errors occurred\n'
                         '====================\n')
            report.write(stderr.decode())
        return (
            TASK_STATUS_OK if shell.returncode in self.ok_status_list else shell.returncode,
            stdout
        )



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.generations main module

This module is checking for a registered tasks scheduler utility.
"""

from pyams_scheduler.interfaces import IScheduler, SCHEDULER_NAME
from pyams_scheduler.scheduler import Scheduler
from pyams_site.generations import check_required_utilities
from pyams_site.interfaces import ISiteGenerations
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


RENAMED_CLASSES = {
    'pyams_scheduler.interfaces ICronTask':
        'pyams_scheduler.interfaces.task ICronTask',
    'pyams_scheduler.interfaces ICronTaskScheduling':
        'pyams_scheduler.interfaces.task ICronTaskScheduling',
    'pyams_scheduler.interfaces IDateTask':
        'pyams_scheduler.interfaces.task IDateTask',
    'pyams_scheduler.interfaces IDateTaskScheduling':
        'pyams_scheduler.interfaces.task IDateTaskScheduling',
    'pyams_scheduler.interfaces ILoopTask':
        'pyams_scheduler.interfaces.task ILoopTask',
    'pyams_scheduler.interfaces ILoopTaskScheduling':
        'pyams_scheduler.interfaces.task ILoopTaskScheduling'
}


REQUIRED_UTILITIES = ((IScheduler, '', Scheduler, SCHEDULER_NAME),)


@utility_config(name='PyAMS scheduler', provides=ISiteGenerations)
class SchedulerGenerationsChecker:
    """Scheduler generations checker"""

    order = 60
    generation = 1

    def evolve(self, site, current=None):  # pylint: disable=unused-argument,no-self-use
        """Check for required utilities"""
        check_required_utilities(site, REQUIRED_UTILITIES)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi module

This module defines base tasks management views.
"""

from pyramid.view import view_config
from zope.interface import Interface

from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_scheduler.interfaces import IBaseTaskScheduling, IScheduler, MANAGE_TASKS_PERMISSION, \
    TASKS_SCHEDULER_LABEL
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.url import absolute_url
from pyams_viewlet.manager import viewletmanager_config
from pyams_zmi.helper.container import delete_container_element
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IMenuHeader, IPropertiesMenu, ISiteManagementMenu
from pyams_zmi.table import I18nColumnMixin, IconColumn, NameColumn, Table, TableAdminView, \
    TableElementEditor, TrashColumn
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@adapter_config(required=(IScheduler, IAdminLayer, Interface, ISiteManagementMenu),
                provides=IMenuHeader)
def scheduler_menu_header(context, request, view, manager):  # pylint: disable=unused-argument
    """Scheduler menu header"""
    return TASKS_SCHEDULER_LABEL


@adapter_config(required=(IScheduler, IAdminLayer, Interface),
                provides=ITableElementEditor)
class SchedulerElementEditor(TableElementEditor):
    """Scheduler element editor"""

    view_name = 'admin'
    modal_target = False

    def __new__(cls, context, request, view):  # pylint: disable=unused-argument
        if not request.has_permission(MANAGE_TASKS_PERMISSION, context=context):
            return None
        return TableElementEditor.__new__(cls)


@viewletmanager_config(name='tasks-list.menu',
                       context=IScheduler, layer=IAdminLayer,
                       manager=ISiteManagementMenu, weight=10,
                       permission=MANAGE_TASKS_PERMISSION,
                       provides=IPropertiesMenu)
class SchedulerTasksListMenu(NavigationMenuItem):
    """Scheduler tasks list menu"""

    label = _("Tasks definition")
    icon_class = 'fa fa-clock'
    href = '#tasks-list.html'


class SchedulerTasksTable(Table):
    """Scheduler tasks table"""

    display_if_empty = True

    @property
    def data_attributes(self):
        attributes = super().data_attributes
        attributes['table'].update({
            'data-ams-order': '2,asc'
        })
        return attributes


@adapter_config(required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IValues)
class SchedulerTasksTableValues(ContextRequestViewAdapter):
    """Scheduler tasks values adapter"""

    @property
    def values(self):
        """Scheduler tasks table values getter"""
        yield from self.context.tasks


@adapter_config(name='active',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskActiveColumn(IconColumn):
    """Scheduler task active column"""

    weight = 1

    def get_icon_hint(self, item):
        translate = self.request.localizer.translate
        return translate(_("Enabled task")) if IBaseTaskScheduling(item).active \
            else translate(_("Disabled task"))

    def get_icon_class(self, item):
        return 'far fa-check-square' if IBaseTaskScheduling(item).active \
            else 'far fa-square text-danger'


@adapter_config(name='icon',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskIconColumn(IconColumn):
    """Scheduler task icon column"""

    weight = 2

    def get_icon_hint(self, item):
        if not item.label:
            return None
        translate = self.request.localizer.translate
        return translate(item.label)

    def get_icon_class(self, item):
        return item.icon_class or None


@adapter_config(name='name',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskNameColumn(NameColumn):
    """Scheduler tasks name column"""


@adapter_config(name='id',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskIdColumn(I18nColumnMixin, GetAttrColumn):
    """Scheduler task ID column"""

    i18n_header = _("Task ID")
    attr_name = '__name__'
    weight = 20


@adapter_config(name='trash',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class SchedulerTaskTrashColumn(TrashColumn):
    """Scheduler tasks trash column"""

    permission = MANAGE_TASKS_PERMISSION


@pagelet_config(name='tasks-list.html', context=IScheduler, layer=IPyAMSLayer,
                permission=MANAGE_TASKS_PERMISSION)
class SchedulerTasksView(TableAdminView):
    """Scheduler tasks view"""

    title = _("Scheduler tasks")
    table_class = SchedulerTasksTable
    table_label = _("List of scheduler tasks")

    @property
    def back_url(self):
        """Form back URL getter"""
        return absolute_url(self.request.root, self.request, 'admin#utilities.html')  # pylint: disable=no-member

    back_url_target = None


@view_config(name='delete-element.json', context=IScheduler, request_type=IPyAMSLayer,
             permission=MANAGE_TASKS_PERMISSION, renderer='json', xhr=True)
def delete_scheduler_task(request):
    """Delete scheduler task"""
    return delete_container_element(request)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.interfaces.task.ssh module

This module defines interface of SSH caller task.
"""

from zope.interface import Interface, Invalid, invariant
from zope.schema import Bool, Int, Object, Password, TextLine

from pyams_scheduler.interfaces import ITask


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


class ISSHConnectionInfo(Interface):
    """SSH connection info interface"""

    hostname = TextLine(title=_("Hostname or IP address"),
                        description=_("Enter hostname or IP address of a remote host; "
                                      "keep empty for local server host"),
                        required=False)

    auto_add_host_key = Bool(title=_("Automatically add host key?"),
                             description=_("If 'yes', connection will be opened event if "
                                           "server key is unknown"),
                             required=False,
                             default=False)

    port = Int(title=_("Port number"),
               default=22,
               required=False)

    username = TextLine(title=_("User name"),
                        required=False)

    private_key = TextLine(title=_("Private key filename"),
                           description=_("Enter name of private key file; use '~' to identify "
                                         "running server user home directory, as in "
                                         "~/.ssh/id_rsa"),
                           required=False)

    password = Password(title=_("Password"),
                        description=_("If not using private key, you must provide user's "
                                      "password"),
                        required=False)

    @invariant
    def check_remote_host(self):
        """Check for password or private key if remote host is defined"""
        if self.hostname and (bool(self.private_key) == bool(self.password)):
            raise Invalid(_("You must provide a private key filename OR a password when "
                            "defining remote tasks"))

    def get_connection(self):
        """Get SSH connection"""

    def get_sftp_client(self):
        """Get SFTP client"""


class ISSHCallTaskInfo(Interface):
    """SSH caller task info"""

    connection = Object(title=_("SSH connection"),
                        schema=ISSHConnectionInfo,
                        required=True)

    cmdline = TextLine(title=_("Command line"),
                       description=_("Enter command line, using absolute path names"),
                       required=True)

    ok_status = TextLine(title=_("OK status"),
                         description=_("Comma-separated list of command exit codes which "
                                       "should be considered as success"),
                         required=True,
                         default='0')


class ISSHCallerTask(ITask, ISSHCallTaskInfo):
    """SSH caller interface"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.jobs module

This module defines components used to display scheduled jobs.
"""

from datetime import datetime

from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_scheduler.interfaces import IScheduler, MANAGE_TASKS_PERMISSION
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.date import format_datetime
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IPropertiesMenu
from pyams_zmi.table import I18nColumnMixin, Table, TableAdminView
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem

__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@viewlet_config(name='jobs-list.menu',
                context=IScheduler, layer=IAdminLayer,
                manager=IPropertiesMenu, weight=20,
                permission=MANAGE_TASKS_PERMISSION)
class SchedulerJobsMenu(NavigationMenuItem):
    """Scheduler jobs menu"""

    label = _("Scheduled jobs")
    href = '#jobs-list.html'


class SchedulerJobsTable(Table):
    """Scheduler jobs table"""


@adapter_config(required=(IScheduler, IAdminLayer, SchedulerJobsTable),
                provides=IValues)
class SchedulerJobsTableValues(ContextRequestViewAdapter):
    """Scheduler jobs values adapter"""

    @property
    def values(self):
        """Scheduler jobs table values getter"""
        status_code, response = self.context.get_jobs()
        if status_code == 200:
            yield from response


class JobsColumn(I18nColumnMixin, GetAttrColumn):
    """Base jobs column"""

    def get_value(self, obj):
        return obj.get(self.attr_name, '--')


@adapter_config(name='name',
                required=(IScheduler, IAdminLayer, SchedulerJobsTable),
                provides=IColumn)
class SchedulerJobsNameColumn(JobsColumn):
    """Scheduler jobs name column"""

    i18n_header = _("Name")
    attr_name = 'name'
    weight = 10


@adapter_config(name='id',
                required=(IScheduler, IAdminLayer, SchedulerJobsTable),
                provides=IColumn)
class SchedulerJobsIDColumn(JobsColumn):
    """Scheduler jobs ID column"""

    i18n_header = _("ID")
    attr_name = 'id'
    weight = 20


@adapter_config(name='trigger',
                required=(IScheduler, IAdminLayer, SchedulerJobsTable),
                provides=IColumn)
class SchedulerJobsTriggerColumn(JobsColumn):
    """Scheduler jobs trigger column"""

    i18n_header = _("Trigger")
    attr_name = 'trigger'
    weight = 30


@adapter_config(name='next_run',
                required=(IScheduler, IAdminLayer, SchedulerJobsTable),
                provides=IColumn)
class SchedulerJobsNextRunColumn(JobsColumn):
    """Scheduler jobs next run column"""

    i18n_header = _("Next run")
    attr_name = 'next_run'
    weight = 40

    def get_value(self, obj):
        value = super().get_value(obj)
        if not value:
            return '--'
        return format_datetime(datetime.utcfromtimestamp(value), request=self.request)


@pagelet_config(name='jobs-list.html', context=IScheduler, layer=IPyAMSLayer,
                permission=MANAGE_TASKS_PERMISSION)
class SchedulerJobsView(TableAdminView):
    """Scheduler jobs view"""

    title = _("Scheduler jobs")
    table_class = SchedulerJobsTable
    table_label = _("List of scheduler jobs")



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.interfaces.task module

This module defines scheduler tasks interfaces.
"""

from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import containers, contains
from zope.container.interfaces import IContainer
from zope.interface import Attribute, Interface
from zope.schema import Bool, Bytes, Choice, Datetime, Float, Int, List, Object, Text, TextLine
from pyams_utils.schema import TextLineListField


__docformat__ = 'restructuredtext'

from pyams_scheduler import _


TASK_STATUS_NONE = None
TASK_STATUS_EMPTY = 'empty'
TASK_STATUS_OK = 'OK'
TASK_STATUS_WARNING = 'warning'
TASK_STATUS_ERROR = 'error'

TASK_STATUS_STYLES = {
    TASK_STATUS_EMPTY: 'table-secondary',
    TASK_STATUS_NONE: 'table-info',
    TASK_STATUS_OK: 'table-success',
    TASK_STATUS_WARNING: 'table-warning',
    TASK_STATUS_ERROR: 'table-danger'
}


#
# Task history item interface
#

class ITaskHistoryContainer(Interface):
    """Task history container interface"""

    contains('pyams_scheduler.interfaces.task.ITaskHistory')


class ITaskHistory(Interface):
    """Scheduler task history item interface"""

    containers(ITaskHistoryContainer)

    status = TextLine(title=_("Execution status"))

    date = Datetime(title=_("Execution start date"),
                    required=True)

    duration = Float(title=_("Execution duration"),
                     required=True)

    report = Text(title=_("Execution report"),
                  required=True)


#
# Scheduler job interface
#

class IJobInfo(Interface):
    """Job interface"""

    id = TextLine(title="Job ID")

    next_run_time = Float(title="Job next run time")

    job_state = Bytes(title="Job state")


#
# Task scheduling modes interfaces
#

TASK_SCHEDULING_MODES_VOCABULARY = 'PyAMS_scheduler.scheduling.modes'


class ITaskSchedulingMode(Interface):
    """Scheduler task scheduling mode"""

    marker_interface = Attribute("Class name of scheduling mode marker interface")

    schema = Attribute("Class name of scheduling mode info interface")

    def get_trigger(self, task):
        """Get trigger for the given task"""

    def schedule(self, task, scheduler):
        """Add given task to the scheduler"""


class ITaskSchedulingMarker(Interface):
    """Base interface for task scheduling mode markers"""


class IBaseTaskScheduling(Interface):
    """Base interface for task scheduling info"""

    active = Bool(title=_("Activate task?"),
                  description=_("You can disable a task by selecting 'No'"),
                  required=True,
                  default=False)

    start_date = Datetime(title=_("First execution date"),
                          description=_("Date and time from which scheduling should start"),
                          required=False)


#
# Scheduler cron-style tasks interfaces
#

SCHEDULER_TASK_CRON_MODE = 'Cron-style scheduling'
SCHEDULER_TASK_CRON_INFO = 'pyams_scheduler.trigger.cron'


class ICronTask(Interface):
    """Cron-style task marker interface"""


class ICronTaskScheduling(IBaseTaskScheduling):
    """Base interface for cron-style scheduled tasks"""

    end_date = Datetime(title=_("Last execution date"),
                        description=_("Date and time past which scheduling should end"),
                        required=False)

    year = TextLine(title=_("Years"),
                    description=_("Years for which to schedule the job"),
                    required=False,
                    default='*')

    month = TextLine(title=_("Months"),
                     description=_("Months (1-12) for which to schedule the job"),
                     required=False,
                     default='*')

    day = TextLine(title=_("Month days"),
                   description=_("Days (1-31) for which to schedule the job"),
                   required=False,
                   default='*')

    week = TextLine(title=_("Weeks"),
                    description=_("Year weeks (1-53) for which to schedule the job"),
                    required=False,
                    default='*')

    day_of_week = TextLine(title=_("Week days"),
                           description=_("Week days (0-6, with 0 as monday) for which to "
                                         "schedule the job"),
                           required=False,
                           default='*')

    hour = TextLine(title=_("Hours"),
                    description=_("Hours (0-23) for which to schedule the job"),
                    required=False,
                    default='*')

    minute = TextLine(title=_("Minutes"),
                      description=_("Minutes (0-59) for which to schedule the job"),
                      required=False,
                      default='*')

    second = TextLine(title=_("Seconds"),
                      description=_("Seconds (0-59) for which to schedule the job"),
                      required=False,
                      default='0')


#
# Scheduler date-style tasks interface
#

SCHEDULER_TASK_DATE_MODE = 'Date-style scheduling'
SCHEDULER_TASK_DATE_INFO = 'pyams_scheduler.trigger.date'


class IDateTask(Interface):
    """Date-style task marker interface"""


class IDateTaskScheduling(IBaseTaskScheduling):
    """Base interface for date-style scheduled tasks"""

    start_date = Datetime(title=_("Execution date"),
                          description=_("Date and time on which execution should start"),
                          required=True)


#
# Scheduler loop-style tasks interface
#

SCHEDULER_TASK_LOOP_MODE = 'Loop-style scheduling'
SCHEDULER_TASK_LOOP_INFO = 'pyams_scheduler.trigger.loop'


class ILoopTask(Interface):
    """Loop-style task marker interface"""


class ILoopTaskScheduling(IBaseTaskScheduling):
    """Base interface for loop-style scheduled tasks"""

    end_date = Datetime(title=_("Last execution date"),
                        description=_("Date and time past which scheduling should end"),
                        required=False)

    weeks = Int(title=_("Weeks interval"),
                description=_("Number of weeks between executions"),
                required=True,
                default=0)

    days = Int(title=_("Days interval"),
               description=_("Number of days between executions"),
               required=True,
               default=0)

    hours = Int(title=_("Hours interval"),
                description=_("Number of hours between executions"),
                required=True,
                default=0)

    minutes = Int(title=_("Minutes interval"),
                  description=_("Number of minutes between executions"),
                  required=True,
                  default=1)

    seconds = Int(title=_("Seconds interval"),
                  description=_("Number of seconds between executions"),
                  required=True,
                  default=0)


#
# Scheduler task interfaces
#

class ITaskInfo(Interface):
    """Scheduler task interface"""

    containers('.IScheduler')

    name = TextLine(title=_("Task name"),
                    description=_("Descriptive name given to this task"),
                    required=True)

    schedule_mode = Choice(title=_("Scheduling mode"),
                           description=_("Scheduling mode defines how task will be scheduled"),
                           vocabulary=TASK_SCHEDULING_MODES_VOCABULARY,
                           required=True)

    keep_empty_reports = Bool(title=_("Keep empty reports history?"),
                              description=_("If 'Yes', empty reports will be kept in task "
                                            "history"),
                              required=True,
                              default=False)

    history_duration = Int(title=_("History duration"),
                           description=_("Number of days during which execution reports are "
                                         "kept in history; enter 0 to remove limit"),
                           required=False)

    history_length = Int(title=_("History max length"),
                         description=_("Number of execution reports to keep in history; enter 0 "
                                       "to remove limit"),
                         required=False)


class ITask(ITaskInfo, IAttributeAnnotatable):
    """Complete task interface"""

    label = Attribute("Task label")

    icon_class = Attribute("FontAwesome icon class")

    settings_view_name = TextLine(title=_("Settings view name"),
                                  default='settings.html',
                                  required=False)

    history = List(title=_("History"),
                   description=_("Task history"),
                   value_type=Object(schema=ITaskHistory))

    runnable = Attribute("Is the task runnable?")

    internal_id = Attribute("Internal ID")

    def get_trigger(self):
        """Get scheduler job trigger"""

    def get_scheduling_info(self):
        """Get scheduling info"""

    def run(self, report, **kwargs):
        """Launch job execution"""

    def store_report(self, report, status, start_date, duration):
        """Store task execution report in task's history"""

    def send_report(self, report, status, target=None):
        """Send task execution report by mail"""

    def reset(self):
        """Re-schedule job execution"""

    def launch(self):
        """Ask task for immediate execution"""


#
# Task notifications interfaces
#

TASK_NOTIFICATION_MODES_VOCABULARY = 'PyAMS_scheduler.notifications.modes'


class ITaskNotificationMode(Interface):
    """Task notification mode utility interface"""

    def send_report(self, task, report, status, target, registry=None):  # pylint: disable=too-many-arguments
        """Send report to given target"""


class ITaskNotification(Interface):
    """Task notification"""

    enabled = Bool(title=_("Enabled?"),
                   description=_("Recipient will not be notified if notification is disabled"),
                   required=True,
                   default=True)

    report_errors_only = Bool(title=_("Only report errors?"),
                              description=_("If 'Yes', only error reports will be sent to given "
                                            "errors target"),
                              required=True,
                              default=False)

    send_empty_reports = Bool(title=_("Send empty reports?"),
                              description=_("If 'No', empty reports will not be sent by mail"),
                              required=True,
                              default=False)

    mode = Attribute("Notification mode")

    def get_handler(self):
        """Return notification handler utility"""


SCHEDULER_TASK_NOTIFICATIONS_KEY = 'pyams_scheduler.task.notifications'


class ITaskNotificationContainer(IContainer):
    """Task notifications container interface"""

    contains(ITaskNotification)

    def get_enabled_items(self):
        """Get iterator over enabled notifications"""


#
# Task mail notifications interfaces
#

class IMailNotification(ITaskNotification):
    """Mail notification info interface"""

    target_email = TextLineListField(title=_("Target's email"),
                                     description=_("Email address of report's recipient; you can "
                                                   "enter email address in a simple form or in "
                                                   "in a complete \"Name <adresse@domain.com>\" "
                                                   "form"),
                                     required=True)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.rest module

This module defines REST caller task.
"""

import codecs
import json
import pprint
from urllib import parse

import chardet
import requests
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces.task import TASK_STATUS_ERROR, TASK_STATUS_OK
from pyams_scheduler.interfaces.task.rest import IRESTCallerTask
from pyams_scheduler.task import Task
from pyams_utils.dict import format_dict
from pyams_utils.factory import factory_config
from pyams_utils.html import html_to_text


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@factory_config(IRESTCallerTask)
class RESTCallerTask(Task):
    """REST API caller task"""

    label = _("HTTP service")
    icon_class = 'fab fa-battle-net'

    base_url = FieldProperty(IRESTCallerTask['base_url'])
    service = FieldProperty(IRESTCallerTask['service'])
    params = FieldProperty(IRESTCallerTask['params'])
    verify_ssl = FieldProperty(IRESTCallerTask['verify_ssl'])
    connection_timeout = FieldProperty(IRESTCallerTask['connection_timeout'])
    allow_redirects = FieldProperty(IRESTCallerTask['allow_redirects'])
    ok_status = FieldProperty(IRESTCallerTask['ok_status'])
    authenticate = FieldProperty(IRESTCallerTask['authenticate'])
    username = FieldProperty(IRESTCallerTask['username'])
    _password = FieldProperty(IRESTCallerTask['password'])
    use_proxy = FieldProperty(IRESTCallerTask['use_proxy'])
    proxy_server = FieldProperty(IRESTCallerTask['proxy_server'])
    proxy_port = FieldProperty(IRESTCallerTask['proxy_port'])
    proxy_username = FieldProperty(IRESTCallerTask['proxy_username'])
    _proxy_password = FieldProperty(IRESTCallerTask['proxy_password'])
    use_jwt_authority = FieldProperty(IRESTCallerTask['use_jwt_authority'])
    jwt_authority_url = FieldProperty(IRESTCallerTask['jwt_authority_url'])
    jwt_token_service = FieldProperty(IRESTCallerTask['jwt_token_service'])
    jwt_login_field = FieldProperty(IRESTCallerTask['jwt_login_field'])
    jwt_password_field = FieldProperty(IRESTCallerTask['jwt_password_field'])
    jwt_token_attribute = FieldProperty(IRESTCallerTask['jwt_token_attribute'])
    jwt_use_proxy = FieldProperty(IRESTCallerTask['jwt_use_proxy'])

    @property
    def password(self):
        """Password getter"""
        return self._password

    @password.setter
    def password(self, value):
        """Password setter"""
        if value == '*****':
            return
        self._password = value

    @property
    def proxy_password(self):
        """Proxy password getter"""
        return self._proxy_password

    @proxy_password.setter
    def proxy_password(self, value):
        """Proxy password setter"""
        if value == '*****':
            return
        self._proxy_password = value

    @property
    def ok_status_list(self):
        """OK status list getter"""
        return map(int, self.ok_status.split(','))

    def run(self, report, **kwargs):
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        # get remote service URL
        method, service = self.service
        rest_service = f'{self.base_url}{service}'
        report.write(f'HTTP service output\n'
                     f'===================\n'
                     f'HTTP service: \n    {method} {rest_service}\n\n')
        # check proxy configuration
        proxies = {}
        if self.use_proxy:
            parsed = parse.urlparse(self.base_url)
            if self.proxy_username:
                proxy_auth = f'{self.proxy_username}:{self.proxy_password}@'
            else:
                proxy_auth = ''
            proxies[parsed.scheme] = f'http://{proxy_auth}{self.proxy_server}:{self.proxy_port}'
        # check JWT authorization
        auth = None
        headers = {}
        if self.use_jwt_authority:
            jwt_method, jwt_service = self.jwt_token_service
            jwt_service = f'{self.jwt_authority_url}{jwt_service}'
            jwt_params = {
                self.jwt_login_field: self.username,
                self.jwt_password_field: self.password
            }
            jwt_request = requests.request(jwt_method, jwt_service,
                                           params=jwt_params if method == 'GET' else None,
                                           data=jwt_params if method != 'GET' else None,
                                           proxies=proxies if self.jwt_use_proxy else None,
                                           timeout=self.connection_timeout,
                                           allow_redirects=False)
            status_code = jwt_request.status_code
            report.write(f'JWT token status code: {status_code}\n')
            if status_code != requests.codes.ok:  # pylint: disable=no-member
                report.write(f'JWT headers: {format_dict(jwt_request.headers)}\n')
                report.write(f'JWT report: {jwt_request.text}\n\n')
                return TASK_STATUS_ERROR, None
            headers['Authorization'] = f'Bearer ' \
                                       f'{jwt_request.json().get(self.jwt_token_attribute)}'
        # build authorization headers
        elif self.username:
            auth = self.username, self.password
        # check params
        params = {}
        if self.params:
            params.update(json.loads(self.params))
        params.update(kwargs)
        # build HTTP request
        rest_request = requests.request(method, rest_service,
                                        auth=auth,
                                        headers=headers,
                                        params=params if method == 'GET' else None,
                                        data=params if method != 'GET' else None,
                                        verify=self.verify_ssl,
                                        proxies=proxies,
                                        timeout=self.connection_timeout,
                                        allow_redirects=self.allow_redirects)
        # check request status
        status_code = rest_request.status_code
        report.write(f'Status code: {status_code}\n')
        report.write(f'Headers: {format_dict(rest_request.headers)}\n\n')
        # check request content
        content_type = rest_request.headers.get('Content-Type', 'text/plain')
        if content_type.startswith('application/json'):
            response = rest_request.json()
            message = pprint.pformat(response)
        elif content_type.startswith('text/html'):
            message = html_to_text(rest_request.text)
        elif content_type.startswith('text/'):
            message = rest_request.text
        else:
            content = rest_request.content
            if 'charset=' in content_type.lower():
                charset = content_type.split('=', 1)[1]
            else:
                charset = chardet.detect(content).get('encoding') or 'utf-8'
            message = codecs.decode(content, charset)
        report.write(message)
        report.write('\n\n')
        return (
            TASK_STATUS_OK if status_code in self.ok_status_list else status_code,
            rest_request
        )
